package de.hft.stuttgart.citydoctor2.healing;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Random;
import java.util.concurrent.atomic.AtomicInteger;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.hft.stuttgart.citydoctor2.check.HealingID;
import de.hft.stuttgart.citydoctor2.check.HealingMethod;
import de.hft.stuttgart.citydoctor2.check.ModificationListener;
import de.hft.stuttgart.citydoctor2.check.error.PolygonWrongOrientationError;
import de.hft.stuttgart.citydoctor2.datastructure.Edge;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;
import de.hft.stuttgart.citydoctor2.utils.Pair;

public class HealPolygonWrongOrientationError implements HealingMethod {

	private static final Logger logger = LogManager.getLogger(HealPolygonWrongOrientationError.class);

	private Random r = new Random();

	@Override
	public boolean visit(PolygonWrongOrientationError err, ModificationListener l) {
		logger.debug("Executing Repair for PolygonWrongOrientationError");
		Map<Polygon, AtomicInteger> polygonCounter = new HashMap<>();
		for (Edge e : err.getEdges()) {
			for (Polygon p : e.getAdjacentPolygons()) {
				AtomicInteger counter = polygonCounter.computeIfAbsent(p, k -> new AtomicInteger(0));
				counter.incrementAndGet();
			}
		}

		LinkedList<Pair<Polygon, AtomicInteger>> data = new LinkedList<>();
		for (Entry<Polygon, AtomicInteger> e : polygonCounter.entrySet()) {
			data.add(new Pair<>(e.getKey(), e.getValue()));
		}
		Collections.sort(data, (o1, o2) -> o2.getValue1().get() - o1.getValue1().get());
		Pair<Polygon, AtomicInteger> item = data.pop();
		// stupid random stuff
		// attempt to avoid creating an endless loop in some cases by reversing the same
		// polygon over and over again
		// choose a polygon randomly from a list, all must have the same number of
		// faulty edges connected to them
		List<Pair<Polygon, AtomicInteger>> samePolygons = new ArrayList<>();
		samePolygons.add(item);
		for (Pair<Polygon, AtomicInteger> temp : data) {
			if (temp.getValue1().get() == item.getValue1().get()) {
				samePolygons.add(temp);
			} else {
				break;
			}
		}

		int index = r.nextInt(samePolygons.size());
		Polygon p = samePolygons.get(index).getValue0();

		Collections.reverse(p.getExteriorRing().getVertices());
		for (LinearRing lr : p.getInnerRings()) {
			Collections.reverse(lr.getVertices());
		}
		return true;
	}

	@Override
	public HealingMethod createNew() {
		return new HealPolygonWrongOrientationError();
	}

	@Override
	public HealingID getID() {
		return HealingID.S_POLYGON_WRONG_ORIENTATION;
	}
}
